import numpy as np
import pandas as pd
from typing import List
import random
from macrosynergy.management.simulate_quantamental_data import make_qdf

from itertools import product


def reduce_df(df: pd.DataFrame, xcats: List[str] = None,  cids: List[str] = None,
              start: str = None, end: str = None, blacklist: dict = None,
              out_all: bool = False, intersect: bool = False):
    """
    Filter DataFrame by xcats and cids and notify about missing xcats and cids.

    :param <pd.Dataframe> df: standardized JPMaQS DataFrame with the necessary columns:
        'cid', 'xcats', 'real_date' and 'value'.
    :param <List[str]> xcats: extended categories to be filtered on. Default is all in
        the DataFrame.
    :param <List[str]> cids: cross sections to be checked on. Default is all in the
        dataframe.
    :param <str> start: string representing the earliest date. Default is None.
    :param <str> end: string representing the latest date. Default is None.
    :param <dict> blacklist: cross-sections with date ranges that should be excluded from
        the data frame. If one cross-section has several blacklist periods append numbers
        to the cross-section code.
    :param <bool> out_all: if True the function returns reduced dataframe and selected/
        available xcats and cids.
        Default is False, i.e. only the DataFrame is returned
    :param <bool> intersect: if True only retains cids that are available for all xcats.
        Default is False.

    :return <pd.Dataframe>: reduced DataFrame that also removes duplicates or
        (for out_all True) DataFrame and available and selected xcats and cids.
    """

    dfx = df[df['real_date'] >= pd.to_datetime(start)] if start is not None else df
    dfx = dfx[dfx['real_date'] <= pd.to_datetime(end)] if end is not None else dfx

    if blacklist is not None:
        for key, value in blacklist.items():
            filt1 = dfx['cid'] == key[:3]
            filt2 = dfx['real_date'] >= pd.to_datetime(value[0])
            filt3 = dfx['real_date'] <= pd.to_datetime(value[1])
            dfx = dfx[~(filt1 & filt2 & filt3)]

    xcats_in_df = dfx['xcat'].unique()
    if xcats is None:
        xcats = sorted(xcats_in_df)
    else:
        xcats = [xcat for xcat in xcats if xcat in xcats_in_df]

    dfx = dfx[dfx['xcat'].isin(xcats)]

    if intersect:
        df_uns = dict(dfx.groupby('xcat')['cid'].unique())
        df_uns = {k: set(v) for k, v in df_uns.items()}
        cids_in_df = list(set.intersection(*list(df_uns.values())))
    else:
        cids_in_df = dfx['cid'].unique()

    if cids is None:
        cids = sorted(cids_in_df)
    else:
        cids = [cids] if isinstance(cids, str) else cids
        cids = [cid for cid in cids if cid in cids_in_df]

        cids = set(cids).intersection(cids_in_df)
        dfx = dfx[dfx['cid'].isin(cids)]

    if out_all:
        return dfx.drop_duplicates(), xcats, sorted(list(cids))
    else:
        return dfx.drop_duplicates()


def reduce_df_by_ticker(df: pd.DataFrame, ticks: List[str] = None,  start: str = None,
                        end: str = None, blacklist: dict = None):
    """
    Filter dataframe by xcats and cids and notify about missing xcats and cids

    :param <pd.Dataframe> df: standardized dataframe with the following columns:
        'cid', 'xcats', 'real_date'.
    :param <List[str]> ticks: tickers (cross sections + base categories)
    :param <str> start: string in ISO 8601 representing earliest date. Default is None.
    :param <str> end: string ISO 8601 representing the latest date. Default is None.
    :param <dict> blacklist: cross sections with date ranges that should be excluded from
        the dataframe. If one cross section has several blacklist periods append numbers
        to the cross section code.

    :return <pd.Dataframe>: reduced dataframe that also removes duplicates
    """

    dfx = df.copy()
    dfx = dfx[dfx["real_date"] >= pd.to_datetime(start)] if start is not None else dfx
    dfx = dfx[dfx["real_date"] <= pd.to_datetime(end)] if end is not None else dfx

    # Blacklisting by cross-section.
    if blacklist is not None:
        for key, value in blacklist.items():
            filt1 = dfx["cid"] == key[:3]
            filt2 = dfx["real_date"] >= pd.to_datetime(value[0])
            filt3 = dfx["real_date"] <= pd.to_datetime(value[1])
            dfx = dfx[~(filt1 & filt2 & filt3)]

    dfx["ticker"] = dfx["cid"] + '_' + dfx["xcat"]
    ticks_in_df = dfx["ticker"].unique()
    if ticks is None:
        ticks = sorted(ticks_in_df)
    else:
        ticks = [tick for tick in ticks if tick in ticks_in_df]

    dfx = dfx[dfx["ticker"].isin(ticks)]

    return dfx.drop_duplicates()


def aggregation_helper(dfx: pd.DataFrame, xcat_agg: str):
    """
    Helper method to down-sample each category in the DataFrame by aggregating over the
    intermediary dates according to a prescribed method.

    :param <List[str]> dfx: standardised DataFrame defined exclusively on a single
        category.
    :param <List[str]> xcat_agg: associated aggregation method for the respective
        category.

    """

    dfx = dfx.groupby(['xcat', 'cid', 'custom_date'])
    dfx = dfx.agg(xcat_agg).reset_index()

    if 'real_date' in dfx.columns:
        dfx = dfx.drop(['real_date'], axis=1)
    dfx = dfx.rename(columns={"custom_date": "real_date"})

    return dfx

def expln_df(df_w: pd.DataFrame, xpls: List[str], agg_meth: str, sum_condition: bool,
             lag: int):
    """
    Produces the explanatory column(s) for the custom DataFrame.

    :param <pd.DataFrame> df_w: group-by DataFrame which has been down-sampled. The
        respective aggregation method will be applied.
    :param <List[str]> xpls: list of explanatory category(s).
    :param <str> agg_meth: aggregation method used for all explanatory variables.
    :param <dict> sum_condition: required boolean to negate erroneous zeros if the
        aggregate method used, for the explanatory variable, is sum.
    :param <int> lag: lag of explanatory category(s). Applied uniformly to each
        category.
    """

    dfw_xpls = pd.DataFrame()
    for xpl in xpls:

        if not sum_condition:
            xpl_col = df_w[xpl].agg(agg_meth).astype(dtype=np.float32)
        else:
            xpl_col = df_w[xpl].sum(min_count=1)

        if lag > 0:
            xpl_col = xpl_col.groupby(level=0).shift(lag)

        dfw_xpls[xpl] = xpl_col

    return dfw_xpls

def categories_df(df: pd.DataFrame, xcats: List[str], cids: List[str] = None,
                  val: str = 'value', start: str = None, end: str = None,
                  blacklist: dict = None, years: int = None, freq: str = 'M',
                  lag: int = 0, fwin: int = 1, xcat_aggs: List[str] = ['mean', 'mean']):

    """
    In principle, create custom two-categories DataFrame with appropriate frequency and,
    if applicable, lags.

    :param <pd.Dataframe> df: standardized JPMaQS DataFrame with the following necessary
        columns: 'cid', 'xcats', 'real_date' and at least one column with values of
        interest.
    :param <List[str]> xcats: extended categories involved in the custom DataFrame. The
        last category in the list represents the dependent variable, and the (n - 1)
        preceding categories will be the explanatory variables(s).
    :param <List[str]> cids: cross-sections to be included. Default is all in the
        DataFrame.
    :param <str> val: name of column that contains the values of interest. Default is
        'value'.
    :param <str> start: earliest date in ISO 8601 format. Default is None,
        i.e. earliest date in DataFrame is used.
    :param <str> end: latest date in ISO 8601 format. Default is None,
        i.e. latest date in DataFrame is used.
    :param <dict> blacklist: cross-sections with date ranges that should be excluded from
        the DataFrame. If one cross section has several blacklist periods append numbers
        to the cross section code.
    :param <int> years: number of years over which data are aggregated. Supersedes the
        "freq" parameter and does not allow lags, Default is None, i.e. no multi-year
        aggregation.
    :param <str> freq: letter denoting frequency at which the series are to be sampled.
        This must be one of 'D', 'W', 'M', 'Q', 'A'. Default is 'M'. Will always be the
        last business day of the respective frequency.
    :param <int> lag: lag (delay of arrival) of explanatory category(s) in periods
        as set by freq. Default is 0.
    :param <int> fwin: forward moving average window of first category. Default is 1,
        i.e no average.
        Note: This parameter is used mainly for target returns as dependent variable.
    :param <List[str]> xcat_aggs: exactly two aggregation methods. Default is 'mean' for
        both. The same aggregation method, the first method in the parameter, will be
        used for all explanatory variables.

    :return <pd.DataFrame>: custom DataFrame with category columns. All rows that contain
        NaNs will be excluded.

    N.B.:
    The number of explanatory categories that can be included is not restricted and will
    be appended column-wise to the returned DataFrame. The order of the DataFrame's
    columns will reflect the order of the categories list.
    """

    frq_options = ['D', 'W', 'M', 'Q', 'A']
    frq_error = f"Frequency parameter must be one of the stated options, {frq_options}."
    assert freq in frq_options, frq_error
    frq_dict = dict(zip(frq_options, ['B', 'W-Fri', 'BM', 'BQ', 'BA']))

    assert isinstance(xcats, list), f"<list> expected and not {type(xcats)}."
    assert all([isinstance(c, str) for c in xcats]), "List of categories expected."
    xcat_error = "The minimum requirement is that a single dependent and explanatory " \
                 "variable are included."
    assert len(xcats) >= 2, xcat_error

    aggs_error = "List of strings, outlining the aggregation methods, expected."
    assert isinstance(xcat_aggs, list), aggs_error
    assert all([isinstance(a, str) for a in xcat_aggs]), aggs_error
    aggs_len = "Only two aggregation methods required. The first will be used for all " \
               "explanatory category(s)."
    assert len(xcat_aggs) == 2, aggs_len

    assert not (years is not None) & (lag != 0), "Lags cannot be applied to year groups."
    if years is not None:
        assert isinstance(start, str), "Year aggregation requires a start date."

        no_xcats = "If the data is aggregated over a multi-year timeframe, only two " \
                   "categories are permitted."
        assert len(xcats) == 2, no_xcats

    df, xcats, cids = reduce_df(df, xcats, cids, start, end, blacklist, out_all=True)

    metric = ["value", "grading", "mop_lag", "eop_lag"]
    val_error = "The column of interest must be one of the defined JPMaQS metrics, " \
                f"{metric}, but received {val}."
    assert val in metric, val_error
    avbl_cols = list(df.columns)
    assert val in avbl_cols, f"The passed column name, {val}, must be present in the " \
                             f"received DataFrame. DataFrame contains {avbl_cols}."

    # Reduce the columns in the DataFrame to the necessary columns:
    # ['cid', 'xcat', 'real_date'] + [val] (name of column that contains the
    # values of interest: "value", "grading", "mop_lag", "eop_lag").
    col_names = ['cid', 'xcat', 'real_date', val]

    df_output = []
    if years is None:

        df_w = df.pivot(index=('cid', 'real_date'), columns='xcat', values=val)

        dep = xcats[-1]
        # The possibility of multiple explanatory variables.
        xpls = xcats[:-1]

        df_w = df_w.groupby([pd.Grouper(level='cid'),
                             pd.Grouper(level='real_date', freq=frq_dict[freq])])

        dfw_xpls = expln_df(
            df_w=df_w, xpls=xpls, agg_meth=xcat_aggs[0],
            sum_condition=(xcat_aggs[0] == "sum"), lag=lag
        )

        # Handles for falsified zeros. Following the frequency conversion, if the
        # aggregation method is set to "sum", time periods that exclusively contain NaN
        # values will incorrectly be summed to the value zero which is misleading for
        # analysis.
        if not (xcat_aggs[-1] == "sum"):
            dep_col = df_w[dep].agg(xcat_aggs[1]).astype(dtype=np.float32)
        else:
            dep_col = df_w[dep].sum(min_count=1)

        if fwin > 1:
            s = 1 - fwin
            dep_col = dep_col.rolling(window=fwin).mean().shift(s)

        dfw_xpls[dep] = dep_col
        # Order such that the return category is the right-most column - will reflect the
        # order of the categories list.
        dfc = dfw_xpls[xpls + [dep]]

    else:
        s_year = pd.to_datetime(start).year
        start_year = s_year
        e_year = df['real_date'].max().year + 1

        grouping = int((e_year - s_year) / years)
        remainder = (e_year - s_year) % years

        year_groups = {}

        for group in range(grouping):
            value = [i for i in range(s_year, s_year + years)]
            key = f"{s_year} - {s_year + (years - 1)}"
            year_groups[key] = value

            s_year += years

        v = [i for i in range(s_year, s_year + (remainder + 1))]
        year_groups[f"{s_year} - now"] = v
        list_y_groups = list(year_groups.keys())

        translate_ = lambda year: list_y_groups[int((year % start_year) / years)]
        df['real_date'] = pd.to_datetime(df['real_date'], errors='coerce')
        df['custom_date'] = df['real_date'].dt.year.apply(translate_)

        dfx_list = [df[df['xcat'] == xcats[0]],
                    df[df['xcat'] == xcats[1]]]
        df_agg = list(map(aggregation_helper, dfx_list, xcat_aggs))
        df_output.extend([d[col_names] for d in df_agg])

        dfc = pd.concat(df_output)
        dfc = dfc.pivot(index=('cid', 'real_date'), columns='xcat',
                        values=val)

    # Adjusted to account for multiple signals requested. If the DataFrame is
    # two-dimensional, signal & a return, NaN values will be handled inside other
    # functionality, as categories_df() is simply a support function. If the parameter
    # how is set to "any", a potential unnecessary loss of data on certain categories
    # could arise.
    return dfc.dropna(axis=0, how='all')


if __name__ == "__main__":

    cids = ['NZD', 'AUD', 'GBP', 'CAD']
    xcats = ['XR', 'CRY', 'GROWTH', 'INFL']
    df_cids = pd.DataFrame(index=cids, columns=['earliest', 'latest', 'mean_add',
                                                'sd_mult'])
    df_cids.loc['AUD'] = ['2000-01-01', '2020-12-31', 0.1, 1]
    df_cids.loc['CAD'] = ['2001-01-01', '2020-11-30', 0, 1]
    df_cids.loc['GBP'] = ['2002-01-01', '2020-11-30', 0, 2]
    df_cids.loc['NZD'] = ['2002-01-01', '2020-09-30', -0.1, 2]

    df_xcats = pd.DataFrame(index=xcats, columns=['earliest', 'latest', 'mean_add',
                                                  'sd_mult', 'ar_coef', 'back_coef'])
    df_xcats.loc['XR'] = ['2000-01-01', '2020-12-31', 0.1, 1, 0, 0.3]
    df_xcats.loc['CRY'] = ['2000-01-01', '2020-10-30', 1, 2, 0.95, 1]
    df_xcats.loc['GROWTH'] = ['2001-01-01', '2020-10-30', 1, 2, 0.9, 1]
    df_xcats.loc['INFL'] = ['2001-01-01', '2020-10-30', 1, 2, 0.8, 0.5]

    black = {'AUD': ['2000-01-01', '2003-12-31'], 'GBP': ['2018-01-01', '2100-01-01']}

    random.seed(2)
    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)

    dfd_x1 = reduce_df(dfd, xcats=xcats[:-1], cids=cids[0],
                       start='2012-01-01', end='2018-01-31')

    tickers = [cid + "_XR" for cid in cids]
    dfd_xt = reduce_df_by_ticker(dfd, ticks=tickers, blacklist=black)

    # Testing categories_df().
    dfc1 = categories_df(
        dfd, xcats=['GROWTH', 'CRY'], cids=cids, val="value", freq='W', lag=1,
        xcat_aggs=['mean', 'mean'], start='2000-01-01', blacklist=black
    )


import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns

from typing import List, Union, Tuple

from macrosynergy.management.simulate_quantamental_data import make_qdf
from macrosynergy.management.shape_dfs import reduce_df


class RealPnL:

    """Estimates and analyses naive illustrative PnLs with limited signal options and
    disregarding transaction costs

    :param <pd.Dataframe> df: standardized DataFrame containing the following columns:
        'cid', 'xcats', 'real_date' and 'value'.
    :param <List[str]> contracts: base tickers (combinations of cross sections and base
        categories) for which signals are to be generated.
    :param <List[str]> sigs: signal categories. These are signal postfixes that are
        appended to the contracts to identify the signal categories
        For example: 'SIG' is appended to 'EUR_FX' to give the ticker 'EUR_FXSIG.
    :param <str> ret: return category postfix; default is "XR_NSA".
    :param <str> start: earliest date in ISO format. Default is None and earliest date
        in df is used.
    :param <str> end: latest date in ISO format. Default is None and latest date in df
        is used.
    :param <dict> blacklist: cross sections with date ranges for which no positions
        should be taken
    """

    def __init__(self, df: pd.DataFrame, contracts: List[str],
                 sigs: List[str], ret: str,
                 cids: List[str] = None,
                 start: str = None, end: str = None,
                 blacklist: dict = None):

        pass

    def make_positions(self):
        """
        Calculate the exact positions for ach contract considering rebalancing frequency
            rebalancing lags, and volaility targets.

        # Todo: define appropriate arguments

        """

    def make_pnls(self):
        """
        Calculate the PnLs under consideration of trading and roll costs.

        # Todo: define appropriate arguments

        """



from .api import Interface
import warnings

warnings.warn(
    "Functionality from macrosynergy.dataquery has been moved to macrosynergy.download.jpmaqs - will be removed in v0.1.0",
    DeprecationWarning,
    stacklevel=2
)

__all__ = ["Interface"]



""" JPMaQS Download Interface """

from typing import List, Optional, Dict, Union
import pandas as pd
import traceback as tb
import datetime
import logging
import warnings
import io
from timeit import default_timer as timer

from macrosynergy.download.dataquery import DataQueryInterface
from macrosynergy.download.exceptions import *
from macrosynergy.management.utils import is_valid_iso_date, Config

logger = logging.getLogger(__name__)
debug_stream_handler = logging.StreamHandler(io.StringIO())
debug_stream_handler.setLevel(logging.NOTSET)
debug_stream_handler.setFormatter(
    logging.Formatter(
        "%(asctime)s - %(levelname)s - %(module)s - %(funcName)s :: %(message)s"
    )
)
logger.addHandler(debug_stream_handler)


class JPMaQSDownload(object):
    """JPMaQS Download Interface Object
    :param <bool> oauth: True if using oauth, False if using username/password with crt/key.

    When using oauth:
    :param <str> client_id: oauth client_id, required if oauth=True.
    :param <str> client_secret: oauth client_secret, required if oauth=True.

    When using username/password with crt/key:
    :param <str> crt: path to crt file.
    :param <str> key: path to key file.
    :param <str> username: username for certificate based authentication.
    :param <str> password : paired with username for certificate.

    When using a config file:
    :param <str> credentials_config: path to config file.

    The config file should contain the client_id and client_secret for oauth, or the
    crt, key, username, and password for certificate based authentication.
    (see macrosynergy.management.utils.JPMaQSAPIConfigObject)

    :param <bool> debug: True if debug mode, False if not.
    :param <bool> suppress_warning: True if suppressing warnings, False if not.
    :param <bool> check_connection: True if the interface should check the connection to
        the server before sending requests, False if not. False by default.

    :param <dict> proxy: proxy to use for requests, None if not using proxy (default).
    :param <bool> print_debug_data: True if debug data should be printed, False if not
        (default).
    :param <dict> dq_kwargs: additional arguments to pass to the DataQuery API object such
        `calender` and `frequency` for the DataQuery API. For more fine-grained usage,
        initialize the DataQueryInterface object explicitly.
    :param <dict> kwargs: additional arguments to pass to the DataQuery API object such as
            :param <str> base_url: base url for the DataQuery API.
            :param <str> calendar: calendar setting to use with the DataQuery API.
            :param <str> frequency: frequency setting to use with the DataQuery API.
            ...
        See macrosynergy.download.dataquery.DataQueryInterface for more.

    :return <JPMaQSDownload>: JPMaQSDownload object

    :raises <TypeError>: if provided arguments are not of the correct type.
    :raises <ValueError>: if provided arguments are invalid or semantically incorrect.

    """

    def __init__(
        self,
        oauth: bool = True,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        crt: Optional[str] = None,
        key: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        check_connection: bool = True,
        credentials_config: Optional[str] = None,
        proxy: Optional[Dict] = None,
        suppress_warning: bool = True,
        debug: bool = False,
        print_debug_data: bool = False,
        dq_download_kwargs: dict = {},
        **kwargs,
    ):
        vars_types_zip: zip = zip(
            [oauth, check_connection, suppress_warning, debug, print_debug_data],
            [
                "oauth",
                "check_connection",
                "suppress_warning",
                "debug",
                "print_debug_data",
            ],
        )
        for varx, namex in vars_types_zip:
            if not isinstance(varx, bool):
                raise TypeError(f"`{namex}` must be a boolean.")

        if not isinstance(proxy, dict) and proxy is not None:
            raise TypeError("`proxy` must be a dictionary or None.")

        if not isinstance(dq_download_kwargs, dict):
            raise TypeError("`dq_download_kwargs` must be a dictionary.")

        self.suppress_warning = suppress_warning
        self.debug = debug
        self.print_debug_data = print_debug_data
        self._check_connection = check_connection
        self.dq_download_kwargs = dq_download_kwargs

        if credentials_config is not None:
            if not isinstance(credentials_config, str):
                raise TypeError("`credentials_config` must be a string.")

        config_obj: Config = Config(
            config_path=credentials_config,
            client_id=client_id,
            client_secret=client_secret,
            crt=crt,
            key=key,
            username=username,
            password=password,
            proxy=proxy,
        )

        self.dq_interface: DataQueryInterface = DataQueryInterface(
            oauth=oauth,
            check_connection=check_connection,
            config=config_obj,
            debug=debug,
            **kwargs,
        )

        self.valid_metrics: List[str] = ["value", "grading", "eop_lag", "mop_lag"]
        self.msg_errors: List[str] = []
        self.msg_warnings: List[str] = []
        self.unavailable_expressions: List[str] = []
        self.downloaded_data: Dict = {}
        self.config_obj: Config = config_obj

        if self._check_connection:
            self.check_connection()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type:
            print(f"Exception: {exc_type} {exc_value}")
            print(tb.print_exc())
            raise exc_type(exc_value)

    @staticmethod
    def construct_expressions(
        tickers: Optional[List[str]] = None,
        cids: Optional[List[str]] = None,
        xcats: Optional[List[str]] = None,
        metrics: Optional[List[str]] = None,
    ) -> List[str]:
        """Construct expressions from the provided arguments.

        :param <list[str]> tickers: list of tickers.
        :param <list[str]> cids: list of cids.
        :param <list[str]> xcats: list of xcats.
        :param <list[str]> metrics: list of metrics.

        :return <list[str]>: list of expressions.
        """

        if tickers is None:
            tickers = []
        if cids is not None and xcats is not None:
            tickers += [f"{cid}_{xcat}" for cid in cids for xcat in xcats]

        return [f"DB(JPMAQS,{tick},{metric})" for tick in tickers for metric in metrics]

    @staticmethod
    def deconstruct_expression(
        expression: Union[str, List[str]]
    ) -> Union[List[str], List[List[str]]]:
        """
        Deconstruct an expression into a list of cid, xcat, and metric.
        Coupled with to JPMaQSDownload.time_series_to_df(), achieves the inverse of
        JPMaQSDownload.construct_expressions().

        :param <str> expression: expression to deconstruct. If a list is provided,
            each element will be deconstructed and returned as a list of lists.

        :return <list[str]>: list of cid, xcat, and metric.

        :raises TypeError: if `expression` is not a string or a list of strings.
        :raises ValueError: if `expression` is an empty list.
        """
        if not isinstance(expression, (str, list)):
            raise TypeError("`expression` must be a string or a list of strings.")

        if isinstance(expression, list):
            if not all(isinstance(exprx, str) for exprx in expression):
                raise TypeError("All elements of `expression` must be strings.")
            elif len(expression) == 0:
                raise ValueError("`expression` must be a non-empty list.")
            return [
                JPMaQSDownload.deconstruct_expression(exprx) for exprx in expression
            ]
        else:
            try:
                exprx: str = expression.replace("DB(JPMAQS,", "").replace(")", "")
                ticker, metric = exprx.split(",")
                result: List[str] = ticker.split("_", 1) + [metric]
                if len(result) != 3:
                    raise ValueError(f"{exprx} is not a valid JPMaQS expression.")
                return ticker.split("_", 1) + [metric]
            except Exception as e:
                warnings.warn(
                    f"Failed to deconstruct expression `{expression}`: {e}",
                    UserWarning,
                )
                # fail safely, return list where all entries are =expression
                return [expression, expression, expression]

    def validate_downloaded_df(
        self,
        data_df: pd.DataFrame,
        expected_expressions: List[str],
        found_expressions: List[str],
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
        verbose: bool = True,
    ) -> bool:
        """
        Validate the downloaded data in the provided dataframe.

        :param <pd.DataFrame> data_df: dataframe containing the downloaded data.
        :param <list[str]> expected_expressions: list of expressions that were expected to be
            downloaded.
        :param <list[str]> found_expressions: list of expressions that were actually downloaded.
        :param <str> start_date: start date of the downloaded data.
        :param <str> end_date: end date of the downloaded data.
        :param <bool> verbose: whether to print the validation results.

        :return <bool>: True if the downloaded data is valid, False otherwise.

        :raises <TypeError>: if `data_df` is not a dataframe.
        """

        if not isinstance(data_df, pd.DataFrame):
            raise TypeError("`data_df` must be a dataframe.")
        if data_df.empty:
            return False

        # check if all expressions are present in the df
        exprs_f = set(found_expressions)
        expr_expected = set(expected_expressions)
        expr_missing = expr_expected - exprs_f
        self.unavailable_expressions += list(expr_missing)
        unexpected_exprs = exprs_f - expr_expected
        if unexpected_exprs:
            raise InvalidDataframeError(
                f"Unexpected expressions were found in the downloaded data: {unexpected_exprs}"
            )

        if expr_missing:
            log_str = (
                f"Some expressions are missing from the downloaded data."
                " Check logger output for complete list. \n"
                f"{len(expr_missing)} out of {len(expr_expected)} expressions are missing."
                f"To download the catalogue of all available expressions and filter the"
                " unavailable expressions, set `get_catalogue=True` in the "
                " call to `JPMaQSDownload.download()`."
            )

            logger.info(log_str)
            logger.info(f"Missing expressions: {expr_missing}")
            if verbose:
                print(log_str)

        # check if all dates are present in the df
        # NOTE : Hardcoded max start date to 1990-01-01. This is because the JPMAQS database
        #        does not have data before this date.
        if datetime.datetime.strptime(
            start_date, "%Y-%m-%d"
        ) < datetime.datetime.strptime("1990-01-01", "%Y-%m-%d"):
            start_date = "1950-01-01"

        dates_expected = pd.bdate_range(start=start_date, end=end_date)
        dates_missing = len(data_df) - len(
            data_df[data_df["real_date"].isin(dates_expected)]
        )

        if dates_missing > 0:
            log_str = (
                f"Some dates are missing from the downloaded data. \n"
                f"{len(dates_missing)} out of {len(dates_expected)} dates are missing."
            )
            logger.warning(log_str)
            if verbose:
                print(log_str)

        # check number of nas in each column
        nas = data_df.isna().sum(axis=0)
        nas = nas[nas > 0]
        if len(nas) > 0:
            log_str = (
                "Some columns have missing values.\n"
                f"Total rows : {len(data_df)} \n"
                "Missing values by column:\n"
                f"{nas.head(10)}"
            )
            logger.warning(log_str)
            if verbose:
                print(log_str)

        return True

    def time_series_to_df(
        self,
        dicts_list: List[Dict],
        validate_df: bool = True,
        expected_expressions: Optional[List[str]] = None,
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
        verbose: bool = True,
    ) -> pd.DataFrame:
        """
        Convert the downloaded data to a pandas DataFrame.
        Parameters
        :param dicts_list <list>: List of dictionaries containing time series
            data from the DataQuery API
        Returns
        :return <pd.DataFrame>: JPMaQS standard dataframe with columns:
            real_date, cid, xcat, <metric>. The <metric> column contains the
            observed data for the given cid and xcat on the given real_date.

        :raises <InvalidDataError>: if the downloaded dataframe is invalid.
        """
        dfs: List[pd.DataFrame] = []
        cid: str
        xcat: str
        found_expressions: List[str] = []
        _missing_exprs: List[str] = []
        self.unavailable_expr_messages = []
        # _missing_exprs is just a sanity check to verify self.unavailable_expressions
        for d in dicts_list:
            cid, xcat, metricx = JPMaQSDownload.deconstruct_expression(
                d["attributes"][0]["expression"]
            )
            if d["attributes"][0]["time-series"] is not None:
                found_expressions.append(d["attributes"][0]["expression"])
                df: pd.DataFrame = (
                    pd.DataFrame(
                        d["attributes"][0]["time-series"],
                        columns=["real_date", metricx],
                    )
                    .assign(cid=cid, xcat=xcat, metric=metricx)
                    .rename(columns={metricx: "obs"})
                )
                df = df[["real_date", "cid", "xcat", "obs", "metric"]]
                dfs.append(df)
            else:
                _missing_exprs.append(d["attributes"][0]["expression"])
                if "message" in d["attributes"][0]:
                    self.unavailable_expr_messages.append(d["attributes"][0]["message"])
                else:
                    self.unavailable_expr_messages.append(
                        f"DataQuery did not return data or error message for expression {d['attributes'][0]['expression']}"
                    )

        assert set(_missing_exprs) == set(
            self.unavailable_expressions
        ), "Downloaded `dicts_list` has been modified before calling `time_series_to_df`"

        if len(dfs) == 0:
            raise InvalidDataframeError(
                "No data was downloaded. Check logger output for complete list of missing expressions."
            )

        final_df: pd.DataFrame = pd.concat(dfs, ignore_index=True)
        dups_df: pd.DataFrame = final_df.groupby(
            ["real_date", "cid", "xcat", "metric"]
        )["obs"].count()
        if sum(dups_df > 1) > 0:
            dups_df = pd.DataFrame(dups_df[dups_df > 1].index.tolist())
            err_str: str = "Duplicate data found for the following expressions:\n"
            for i in dups_df.groupby([1, 2, 3]).groups:
                dts_series: pd.Series = dups_df.iloc[
                    dups_df.groupby([1, 2, 3]).groups[i]
                ][0]
                dts: List[str] = dts_series.tolist()
                max_date: str = pd.to_datetime(max(dts)).strftime("%Y-%m-%d")
                min_date: str = pd.to_datetime(min(dts)).strftime("%Y-%m-%d")
                expression: str = self.construct_expressions(
                    cids=[i[0]], xcats=[i[1]], metrics=[i[2]]
                )[0]
                err_str += (
                    f"Expression: {expression}, Dates: {min_date} to {max_date}\n"
                )

            raise InvalidDataframeError(err_str)

        final_df = (
            final_df.set_index(["real_date", "cid", "xcat", "metric"])["obs"]
            .unstack(3)
            .rename_axis(None, axis=1)
            .reset_index()
        )  # thank you @mikiinterfiore

        final_df["real_date"] = pd.to_datetime(final_df["real_date"])

        # list expected business dates, and non-business dates
        expc_bdates: pd.DatetimeIndex = pd.bdate_range(start=start_date, end=end_date)
        expc_nbdates: pd.DatetimeIndex = pd.DatetimeIndex(
            list(set(pd.date_range(start=start_date, end=end_date)) - set(expc_bdates))
        )

        # check if any dates in the downloaded data are not in the expected dates (business + non-business)
        dates_bools: pd.Series = final_df["real_date"].isin(expc_bdates) | final_df[
            "real_date"
        ].isin(expc_nbdates)

        unexpected_dates: pd.DatetimeIndex = final_df[~dates_bools][
            "real_date"
        ].unique()

        if any(~dates_bools):
            raise InvalidDataframeError(
                f"Unexpected dates were found in the downloaded data: {unexpected_dates}"
            )

        # finally, filter out the non-business dates
        final_df = final_df[final_df["real_date"].isin(expc_bdates)]

        final_df = final_df.sort_values(["real_date", "cid", "xcat"])

        found_metrics = sorted(
            list(set(final_df.columns) - {"real_date", "cid", "xcat"}),
            key=lambda x: self.valid_metrics.index(x),
        )
        # sort found_metrics in the order of self.valid_metrics, then re-order the columns
        final_df = final_df[["real_date", "cid", "xcat"] + found_metrics]

        # IMPORTANT NOTE:
        # Drop NA containing rows to be revisited when blacklisting is implemented

        final_df = final_df.dropna(axis=0, how="any").reset_index(drop=True)

        if validate_df:
            vdf = self.validate_downloaded_df(
                data_df=final_df,
                expected_expressions=expected_expressions,
                found_expressions=found_expressions,
                start_date=start_date,
                end_date=end_date,
                verbose=verbose,
            )
            if not vdf:
                self.downloaded_data: Dict = {
                    "data_df": final_df,
                    "expected_expressions": expected_expressions,
                    "found_expressions": found_expressions,
                    "start_date": start_date,
                    "end_date": end_date,
                    "downloaded_dicts": dicts_list,
                }

                raise InvalidDataframeError(
                    "The downloaded data is not valid. "
                    "Please check JPMaQSDownload.downloaded_data "
                    "(dict) for more information."
                )

        return final_df

    def check_connection(
        self, verbose: bool = False, raise_error: bool = False
    ) -> bool:
        """Check if the interface is connected to the server.
        :return <bool>: True if connected, False if not.
        """

        res = self.dq_interface.check_connection(verbose=verbose)
        if raise_error and not res:
            raise ConnectionError(HeartbeatError("Heartbeat failed."))
        return res

    def validate_download_args(
        self,
        tickers: List[str],
        cids: List[str],
        xcats: List[str],
        metrics: List[str],
        start_date: str,
        end_date: str,
        get_catalogue: bool,
        expressions: List[str],
        show_progress: bool,
        as_dataframe: bool,
        report_time_taken: bool,
        report_egress: bool,
    ) -> bool:
        """Validate the arguments passed to the download function.

        :params -- see macrosynergy.download.jpmaqs.JPMaQSDownload.download()

        :return <bool>: True if valid.

        :raises <TypeError>: If any of the arguments are not of the correct type.
        :raises <ValueError>: If any of the arguments are semantically incorrect.

        """

        if not isinstance(show_progress, bool):
            raise TypeError("`show_progress` must be a boolean.")

        if not isinstance(as_dataframe, bool):
            raise TypeError("`as_dataframe` must be a boolean.")

        if not isinstance(report_time_taken, bool):
            raise TypeError("`report_time_taken` must be a boolean.")

        if not isinstance(report_egress, bool):
            raise TypeError("`report_egress` must be a boolean.")

        if not isinstance(get_catalogue, bool):
            raise TypeError("`get_catalogue` must be a boolean.")

        if all([tickers is None, cids is None, xcats is None, expressions is None]):
            raise ValueError(
                "Must provide at least one of `tickers`, "
                "`expressions`, or `cids` & `xcats` together."
            )
        vars_types_zip: zip = zip(
            [tickers, cids, xcats, expressions],
            ["tickers", "cids", "xcats", "expressions"],
        )
        for varx, namex in vars_types_zip:
            if not isinstance(varx, list) and varx is not None:
                raise TypeError(f"`{namex}` must be a list of strings.")
            if varx is not None:
                if len(varx) > 0:
                    if not all([isinstance(ticker, str) for ticker in varx]):
                        raise TypeError(f"`{namex}` must be a list of strings.")
                else:
                    raise ValueError(f"`{namex}` must be a non-empty list of strings.")

        if metrics is None:
            raise ValueError("`metrics` must be a non-empty list of strings.")
        else:
            if all([metric not in self.valid_metrics for metric in metrics]):
                raise ValueError(f"`metrics` must be a subset of {self.valid_metrics}.")

        if cids is not None:
            if xcats is None:
                raise ValueError(
                    "If specifying `cids`, `xcats` must also be specified."
                )
        else:
            if xcats is not None:
                raise ValueError(
                    "If specifying `xcats`, `cids` must also be specified."
                )

        for varx, namex in zip([start_date, end_date], ["start_date", "end_date"]):
            if not isinstance(varx, str):
                raise TypeError(f"`{namex}` must be a string.")
            if not is_valid_iso_date(varx):
                raise ValueError(
                    f"`{namex}` must be a valid date in the format YYYY-MM-DD."
                )
            if pd.to_datetime(varx, errors="coerce") is pd.NaT:
                raise ValueError(
                    f"`{namex}` must be a valid date > "
                    f"{pd.Timestamp.min.strftime('%Y-%m-%d')} "
                )
            if pd.to_datetime(varx) < pd.to_datetime("1950-01-01"):
                warnings.warn(
                    message=(
                        f"`{namex}` is set before 1950-01-01."
                        "Data before 1950-01-01 may not be available,"
                        " and will cause errors/missing data."
                    ),
                    category=UserWarning,
                )

        return True

    def get_catalogue(self):
        self.catalogue: List[str] = self.dq_interface.get_catalogue()
        return self.catalogue

    def filter_expressions_from_catalogue(
        self, expressions: List[str], verbose: bool = True
    ) -> List[str]:
        """
        Method to filter a list of expressions against the JPMaQS catalogue.
        This avoids requesting data for expressions that are not in the catalogue,
        and provides the user wuth the complete list of expressions that are in the
        catalogue.

        Parameters
        :param <List[str]> tickers: list of tickers to filter.

        :return <List[str]>: list of tickers that are in the JPMaQS catalogue.
        """
        print("Downloading the JPMaQS catalogue from DataQuery...")
        catalogue_tickers: List[str] = self.get_catalogue()
        catalogue_expressions: List[str] = self.construct_expressions(
            tickers=catalogue_tickers, metrics=self.valid_metrics
        )
        r: List[str] = sorted(
            list(set(expressions).intersection(catalogue_expressions))
        )
        if verbose:
            filtered: int = len(expressions) - len(r)
            if filtered > 0:
                print(
                    f"Removed {filtered}/{len(expressions)} expressions that are not in the JPMaQS catalogue."
                )

        return r

    def download(
        self,
        tickers: Optional[List[str]] = None,
        cids: Optional[List[str]] = None,
        xcats: Optional[List[str]] = None,
        metrics: List[str] = ["value"],
        start_date: str = "2000-01-01",
        end_date: Optional[str] = None,
        expressions: Optional[List[str]] = None,
        get_catalogue: bool = False,
        show_progress: bool = False,
        debug: bool = False,
        suppress_warning: bool = False,
        as_dataframe: bool = True,
        report_time_taken: bool = False,
        report_egress: bool = False,
    ) -> Union[pd.DataFrame, List[Dict]]:
        """Driver function to download data from JPMaQS via the DataQuery API.
        Timeseries data can be requested using `tickers` with `metrics`, or
        passing formed DataQuery expressions.
        `cids` and `xcats` (along with `metrics`) are used to construct
        expressions, which are ultimately passed to the DataQuery Interface.

        :param <list[str]> tickers: list of tickers.
        :param <list[str]> cids: list of cids.
        :param <list[str]> xcats: list of xcats.
        :param <list[str]> metrics: list of metrics, one of "value" (default),
            "grading", "eop_lag", "mop_lag". "all" is also accepted.
        :param <str> start_date: start date of the data to download, in the
            ISO format - YYYY-MM-DD.
        :param <str> end_date: end date of the data to download in the ISO
            format - YYYY-MM-DD.
        :param <list[str]> expressions: list of DataQuery expressions.
        :param <bool> get_catalogue: If True, the JPMaQS catalogue is
            downloaded and used to filter the list of tickers. Default is
            False.
        :param <bool> show_progress: True if progress bar should be shown,
            False if not (default).
        :param <bool> suppress_warning: True if suppressing warnings. Default
            is True.
        :param <bool> debug: Override the debug behaviour of the JPMaQSDownload
            class. If True, debug mode is enabled.
        :param <bool> print_debug_data: True if debug data should be printed,
            False if not (default). If debug=True, this is set to True.
        :param <bool> as_dataframe: Return a dataframe if True (default),
            a list of dictionaries if False.
        :param <bool> report_time_taken: If True, the time taken to download
            and apply data transformations is reported.
        :param <bool> report_egress: If True, the number of bytes downloaded
            is reported along with the transmission speed in kilobits/second.

        :return <pd.DataFrame|list[Dict]>: dataframe of data if
            `as_dataframe` is True, list of dictionaries if False.

        :raises <ValueError>: if provided arguments are invalid or
            semantically incorrect (see
            macrosynergy.download.jpmaqs.JPMaQSDownload.validate_download_args()).

        """

        # override the default warning behaviour and debug behaviour
        self.suppress_warning = suppress_warning
        self.debug = debug

        vartolist = lambda x: [x] if isinstance(x, str) else x
        tickers = vartolist(tickers)
        cids = vartolist(cids)
        xcats = vartolist(xcats)
        expressions = vartolist(expressions)
        metrics = vartolist(metrics)

        if len(metrics) == 1:
            if metrics[0] == "all":
                metrics = self.valid_metrics

        if end_date is None:
            end_date = (datetime.datetime.today() + pd.offsets.BusinessDay(2)).strftime(
                "%Y-%m-%d"
            )
            # NOTE : due to timezone conflicts, we choose to request data for 2 days in the future.
            # NOTE : DataQuery specifies YYYYMMDD as the date format, but we use YYYY-MM-DD for consistency.
            #   This is date is cast to YYYYMMDD in macrosynergy.download.dataquery.py.

        # Validate arguments.
        if not self.validate_download_args(
            tickers=tickers,
            cids=cids,
            xcats=xcats,
            metrics=metrics,
            start_date=start_date,
            end_date=end_date,
            expressions=expressions,
            get_catalogue=get_catalogue,
            show_progress=show_progress,
            as_dataframe=as_dataframe,
            report_time_taken=report_time_taken,
            report_egress=report_egress,
        ):
            raise ValueError("Invalid arguments passed to download().")
        if pd.to_datetime(start_date) > pd.to_datetime(end_date):
            warnings.warn(
                message=(
                    f"`start_date` ({start_date}) is after `end_date` ({end_date}). "
                    "These dates will be swapped."
                ),
                category=UserWarning,
            )
            start_date, end_date = end_date, start_date

        # Construct expressions.
        if expressions is None:
            expressions: List[str] = []

        expressions += self.construct_expressions(
            tickers=tickers,
            cids=cids,
            xcats=xcats,
            metrics=metrics,
        )
        expressions = list(set(expressions))

        if get_catalogue:
            expressions = self.filter_expressions_from_catalogue(expressions)

        # Download data.
        data: List[Dict] = []
        egress_data: Dict = {}
        download_time_taken: float = timer()
        with self.dq_interface as dq:
            print(
                "Downloading data from JPMaQS.\nTimestamp UTC: ",
                datetime.datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S"),
            )
            self.dq_interface.check_connection(verbose=True)
            print(f"Number of expressions requested: {len(expressions)}")
            data = dq.download_data(
                expressions=expressions,
                start_date=start_date,
                end_date=end_date,
                show_progress=show_progress,
                **self.dq_download_kwargs,
            )

            if len(self.dq_interface.msg_errors) > 0:
                self.msg_errors += self.dq_interface.msg_errors

            if len(self.dq_interface.msg_warnings) > 0:
                self.msg_warnings += self.dq_interface.msg_warnings

            if len(self.dq_interface.unavailable_expressions) > 0:
                self.unavailable_expressions += (
                    self.dq_interface.unavailable_expressions
                )

            if report_egress:
                egress_data = self.dq_interface.egress_data

        download_time_taken: float = timer() - download_time_taken
        dfs_time_taken: float = timer()
        if as_dataframe:
            data_df: pd.DataFrame = self.time_series_to_df(
                dicts_list=data,
                validate_df=True,
                expected_expressions=expressions,
                start_date=start_date,
                end_date=end_date,
                verbose=not self.suppress_warning,
            )
            data = data_df

        dfs_time_taken: float = timer() - dfs_time_taken

        if report_time_taken:
            print(f"Time taken to download data: \t{download_time_taken:.2f} seconds.")
            if as_dataframe:
                print(
                    f"Time taken to convert to dataframe: \t{dfs_time_taken:.2f} seconds."
                )

        if report_egress:
            # create averages for egress_data like
            #  egress_data[tracking_id] = {
            #     "url": log_url,
            #     "upload_size": upload_size,
            #     "download_size": download_size,
            #     "time_taken": time_taken,
            #   }

            total_upload: int = 0
            total_download: int = 0
            total_time_taken: float = 0
            longest_time_taken: float = 0
            longest_time_taken_url: str = ""
            for tracking_id in egress_data:
                total_upload += egress_data[tracking_id]["upload_size"]
                total_download += egress_data[tracking_id]["download_size"]
                total_time_taken += egress_data[tracking_id]["time_taken"]
                if egress_data[tracking_id]["time_taken"] > longest_time_taken:
                    longest_time_taken = egress_data[tracking_id]["time_taken"]
                    longest_time_taken_url = egress_data[tracking_id]["url"]

            avg_upload_size_kb: float = total_upload / (1024)
            avg_download_size_kb: float = total_download / (1024)
            avg_time_taken: float = total_time_taken / len(egress_data)
            avg_transfer_rate_kbit: float = (
                (avg_download_size_kb + avg_upload_size_kb) * 8 / avg_time_taken
            )
            print(f"Average upload size: \t{avg_upload_size_kb:.2f} KB")
            print(f"Average download size: \t{avg_download_size_kb:.2f} KB")
            print(f"Average time taken: \t{avg_time_taken:.2f} seconds")
            print(f"Longest time taken: \t{longest_time_taken:.2f} seconds")
            print(f"Average transfer rate : \t{avg_transfer_rate_kbit:.2f} Kbps")

        if len(self.msg_errors) > 0:
            if not self.suppress_warning:
                print(
                    f"{len(self.msg_errors)} errors encountered during the download. \n"
                    f"The errors did not compromise the download. \n"
                    f"Please check `JPMaQSDownload.msg_errors` for more information."
                )

        return data


if __name__ == "__main__":
    cids = [
        "AUD",
        "BRL",
        "CAD",
        "CHF",
        "CLP",
        "CNY",
        "COP",
        "CZK",
        "DEM",
        "ESP",
        "EUR",
        "FRF",
        "GBP",
        "USD",
    ]
    xcats = [
        "RIR_NSA",
        "FXXR_NSA",
        "FXXR_VT10",
        "DU05YXR_NSA",
        "DU05YXR_VT10",
    ]
    metrics = "all"
    start_date: str = "2023-01-01"
    end_date: str = "2023-03-20"

    with JPMaQSDownload(
        credentials_config="env",
        debug=True,
    ) as jpmaqs:
        data = jpmaqs.download(
            cids=cids,
            xcats=xcats,
            metrics=metrics,
            start_date=start_date,
            get_catalogue=True,
            end_date=end_date,
            show_progress=True,
            suppress_warning=False,
            report_time_taken=True,
            report_egress=True,
        )

        print(data.head())



"""Macrosynergy Quant Research Package"""
import os
import sys
import subprocess
import warnings
from pathlib import Path

DOCLINES = (__doc__ or '').split("\n")

path = Path(__file__).parent
with open(os.path.join(path, "README.md"), "r") as f:
    readme = f.read()

if sys.version_info[:2] < (3, 6):
    raise RuntimeError("Python version >= 3.6 required.")

CLASSIFIERS = """\
Intended Audience :: Science/Research
Intended Audience :: Developers
Programming Language :: Python
Programming Language :: Python :: 3
Programming Language :: Python :: 3.8
Programming Language :: Python :: 3.9
Programming Language :: Python :: 3.10
Programming Language :: Python :: 3.11
Programming Language :: Python :: 3 :: Only
Topic :: Software Development
Topic :: Scientific/Engineering
Typing :: Typed
Operating System :: Microsoft :: Windows
Operating System :: POSIX
Operating System :: MacOS
Development Status :: 4 - Beta
"""

MAJOR = 0
MINOR = 0
MICRO = 32
ISRELEASED = True
VERSION = '%d.%d.%d' % (MAJOR, MINOR, MICRO)

if sys.version_info >= (3, 12):
    # The first version not in the `Programming Language :: Python :: ...` classifiers above
    warnings.warn(
        f"Macrosynergy {VERSION} may not yet support Python "
        f"{sys.version_info.major}.{sys.version_info.minor}.",
        RuntimeWarning,
    )
    
if sys.version_info < (3, 8):
    warnings.warn(
        f"Python {sys.version_info.major}.{sys.version_info.minor} "
        "has reached end-of-life. The Macrosynergy package no longer supports this version. "
        "Please upgrade to Python 3.8 or later.", 
        RuntimeWarning,
    )


# Return the git revision as a string
def git_version():
    def _minimal_ext_cmd(cmd):
        # construct minimal environment
        env = {}
        for k in ['SYSTEMROOT', 'PATH', 'HOME']:
            v = os.environ.get(k)
            if v is not None:
                env[k] = v
        # LANGUAGE is used on win32
        env['LANGUAGE'] = 'C'
        env['LANG'] = 'C'
        env['LC_ALL'] = 'C'
        out = subprocess.check_output(cmd, stderr=subprocess.STDOUT, env=env)
        return out

    try:
        out = _minimal_ext_cmd(['git', 'rev-parse', 'HEAD'])
        GIT_REVISION = out.strip().decode('ascii')
    except (subprocess.SubprocessError, OSError):
        GIT_REVISION = "Unknown"

    if not GIT_REVISION:
        # this shouldn't happen but apparently can (see gh-8512)
        GIT_REVISION = "Unknown"

    return GIT_REVISION


def get_version_info():
    # Adding the git rev number needs to be done inside write_version_py(),
    # otherwise the import of qstools.version messes up the build under Python 3.
    FULLVERSION = VERSION
    if os.path.exists('.git'):
        GIT_REVISION = git_version()
    elif os.path.exists('macrosynergy/version.py'):
        # must be a source distribution, use existing version file
        try:
            from qstools.version import git_revision as GIT_REVISION
        except ImportError:
            raise ImportError("Unable to import git_revision. Try removing "
                              "qstools/version.py and the build directory "
                              "before building.")
    else:
        GIT_REVISION = "Unknown"

    if not ISRELEASED:
        import time

        time_stamp = time.strftime("%Y%m%d%H%M%S", time.localtime())
        FULLVERSION += f'.dev0+{time_stamp}_{GIT_REVISION[:7]}'

    return FULLVERSION, GIT_REVISION


def write_version_py(filename='macrosynergy/version.py'):
    cnt = """\
# THIS FILE IS GENERATED FROM macrosynergy SETUP.PY
short_version: str = '%(version)s'
version: str = '%(version)s'
full_version: str = '%(full_version)s'
git_revision: str = '%(git_revision)s'
release: bool = %(isrelease)s
if not release:
    version = full_version
"""
    FULLVERSION, GIT_REVISION = get_version_info()

    a = open(filename, 'w')
    try:
        a.write(cnt % {'version': VERSION,
                       'full_version': FULLVERSION,
                       'git_revision': GIT_REVISION,
                       'isrelease': str(ISRELEASED)})
    finally:
        a.close()


with open(os.path.join(os.path.dirname(__file__), "requirements.txt")) as f:
    REQUIREMENTS = f.read()


def setup_package():
    from setuptools import setup, find_packages

    src_path = os.path.dirname(os.path.abspath(__file__))
    old_path = os.getcwd()
    os.chdir(src_path)
    sys.path.insert(0, src_path)

    # Rewrite the version file every time
    write_version_py()

    metadata = dict(
        name='macrosynergy',
        maintainer="Macrosynergy",
        maintainer_email="info@macrosynergy.com",
        description=DOCLINES[0],
        license="MIT License",
        long_description_content_type="text/markdown",
        long_description=readme,
        url="https://www.macrosynergy.com",
        author_email='info@macrosynergy.com',
        author='Macrosynergy Ltd',
        download_url="https://github.com/macrosynergy/macrosynergy",
        project_urls={
            "Bug Tracker": "https://github.com/macrosynergy/macrosynergy/issues",
            "Source Code": "https://github.com/macrosynergy/macrosynergy",
            "Documentation": "https://docs.macrosynergy.com"
        },
        classifiers=[_f for _f in CLASSIFIERS.split('\n') if _f],
        platforms=["Windows", "Linux", "Mac OS-X"],
        test_suite='pytest',
        python_requires='>=3.6',
        install_requires=REQUIREMENTS.split("\n"),
        include_package_data=True,
        packages=find_packages(),
        version=get_version_info()[0],
    )
    # __copyright__ = 'Copyright 2020 Macrosynergy Ltd'

    try:
        setup(**metadata)
    finally:
        del sys.path[0]
        os.chdir(old_path)
    return


if __name__ == "__main__":
    setup_package()



"""Old DataQuery Interface path"""
import warnings
from macrosynergy.dataquery.api import Interface as DataQueryInterface

warnings.warn(
    "DataQueryInterface has been moved to macrosynergy.dataquery.api - will be removed in v0.1.0",
    DeprecationWarning,
    stacklevel=2
)

__all__ = ["DataQueryInterface"]



from .shape_dfs import reduce_df, categories_df, reduce_df_by_ticker
from .check_availability import check_availability, visual_paneldates, check_enddates, check_startyears \
    , missing_in_df, business_day_dif
from .simulate_vintage_data import VintageData
from .simulate_quantamental_data import make_qdf
from .update_df import update_df
from .utils import (
    common_cids,
    Config 
)
from . import utils

__all__ = ['check_availability', 'visual_paneldates', 'check_enddates', 'check_startyears',
           'reduce_df', 'reduce_df_by_ticker', 'missing_in_df', 'VintageData', 'make_qdf',
            'update_df', 'business_day_dif', 'categories_df', 'common_cids', 'Config', 'utils']


import requests


class ExceptionAdapter(Exception):
    """Base class for all exceptions raised by the macrosynergy package."""

    def __init__(self, message: str = ""):
        self.message = message

    def __str__(self):
        # print type and message
        return f"{self.__class__.__name__}: {self.message} \n{self.__class__.__name__}"


class AuthenticationError(ExceptionAdapter):
    """Raised when authentication fails."""


class DownloadError(ExceptionAdapter):
    """Raised when a download fails or is incomplete."""


class InvalidResponseError(ExceptionAdapter):
    """Raised when a response is not valid."""


class HeartbeatError(ExceptionAdapter):
    """Raised when a heartbeat fails."""


class InvalidDataframeError(ExceptionAdapter):
    """Raised when a dataframe is not valid."""


class MissingDataError(ExceptionAdapter):
    """Raised when data is missing from a requested dataframe."""


class NoContentError(ExceptionAdapter):
    """Raised when no data is returned from a request."""


KNOWN_EXCEPTIONS = [
    requests.exceptions.ConnectionError,
    requests.exceptions.ConnectTimeout,
    requests.exceptions.ReadTimeout,
    ConnectionResetError,
    requests.exceptions.Timeout,
    requests.exceptions.TooManyRedirects,
    requests.exceptions.RequestException,
    requests.exceptions.HTTPError,
    requests.exceptions.InvalidURL,
    requests.exceptions.InvalidSchema,
    requests.exceptions.ChunkedEncodingError,
]
